/* 
 * Code generation machinery required to compile DPF trie.
 */
#include <dpf-internal.h>
#include <stddef.h>
#include <stdlib.h>
#include <vcode/vcode.h>

/* We make these globals since passing as params makes more bugs. */

/* Argument registers */
static v_reg_t msg_r;			/* Register holding message ptr */
static v_reg_t nbytes_r;		/* Register holding # of bytes in msg */

/* Used by dpf.  Should just make statically allocated (using T0 & co.). */
static v_reg_t hte_r, val_r;   /* Used by the hash table */
static v_reg_t src_r;			/* Value loaded from message */
static v_reg_t t0;			/* Volatile temp */
static v_reg_t fail_r;			/* Holds pc to jump to on failure. */
static v_reg_t jump_r;			/* Holds pc last jumped to.  */

static int fail_save_area;


/* 
 * Stack offsets used to save and restore the current message and nbytes
 * values on every indirect shift. 
 */
static int nbytes_save_area;
#	define NBYTES_OFFSET(l) ((l) * sizeof(unsigned) + nbytes_save_area)
static int msg_save_area;
#	define MSG_OFFSET(l) ((l) * sizeof(void *) + msg_save_area)

/* To protect against seg faults, dpf-iptr is set to this instead of null */
static int dpf_ret_zero(uint8 *msg, unsigned nbytes) { return 0; }

/* This is the pointer used to access the generated code */
int (*dpf_iptr)(uint8 *msg, unsigned nbytes) = dpf_ret_zero;

/* Compute log base 2 of x. */
unsigned log2(unsigned x) {
	unsigned i;

	demand(x && x == (x & -x), x is not a power of two!);
	for(i = 0; x; x >>= 1, i++) 
		;
	return i-1;
}

/* If parent has two or more kids, a is in an orlist. */
inline int orlist_p(Atom a) {
        Atom firstor = a->parent->kids.lh_first;
        return firstor && firstor->sibs.le_next;
}

/* is a the first or in an orlist? */
static inline int firstor_p(Atom a) {
        return orlist_p(a) && a == a->parent->kids.lh_first;
}

/* Save state modified by shift. */
static void save_shift_state(int level) {
	v_stui(nbytes_r, v_lp, NBYTES_OFFSET(level));
	v_stpi(msg_r, v_lp, MSG_OFFSET(level));
}

/* Restore state of previous shift. */
static void restore_shift_state(int level) {
	v_raw_load(v_ldui(nbytes_r, v_lp, NBYTES_OFFSET(level)), 1);
	v_ldpi(msg_r, v_lp, MSG_OFFSET(level));
}


/*
 * If we have shifted on this path or the offset surpasses
 * the minimal amount of buffer space we are allocated,
 * emit a check to see if we have exceeded our memory.
 */
static void emit_bounds_check(int shiftp, uint32 offset, v_label_t l) {
        if(shiftp || offset > DPF_MINMSG)
                v_bleui(nbytes_r, offset, l);
}

/* Compile the lhs of a message ld: (msg[offset:nbits] &  mask) */
static void compile_msgld(struct ir *ir, v_label_t l) {
	uint32 mask;
	uint16 offset;
	struct eq *e;

	e  = &ir->u.eq;
	mask = e->mask;
	offset = e->offset;
	emit_bounds_check(ir->shiftp, offset, l);

	switch(e->nbits) {
	case 8:		
		v_alduci(src_r, msg_r, offset, ir->alignment);
		if(mask != 0xff)
			v_andui(src_r, src_r, mask); 
		break;
	case 16:
		v_aldusi(src_r, msg_r, offset, ir->alignment);
		if(mask != 0xffff)
			v_andui(src_r, src_r, mask); 
		break;
	case 24:
		v_aldui(src_r, msg_r, offset, ir->alignment);
		v_andui(src_r, src_r, mask);
		break;
	case 32:
		v_aldui(src_r, msg_r, offset, ir->alignment);
		if(mask != 0xffffffff)
			v_andui(src_r, src_r, mask); 
		break;
	/* 
	 * Should extend to 64 bits at least.  Possibly further and use
	 * branchless code for equality testing.
	 */

	/* Death. */
	default: fatal(bogus number of bits);
	}
}

/* Useful offsets into the atom structure. */
#define CODE_OFFSET 	offsetof(struct atom, code)
#define KID_OFFSET   	offsetof(struct atom, kids.lh_first)
#define SIB_OFFSET   	offsetof(struct atom, sibs.le_next)
#define KID_FROM_CODE 	(KID_OFFSET - CODE_OFFSET)
#define SIB_FROM_CODE 	(SIB_OFFSET - CODE_OFFSET)

/* Offset into vector to save area */
#define FAIL_OFFSET(level) ( (level) * sizeof(void *) + fail_save_area)

static void compile_shift(Atom a, v_label_t false) {
	uint8 shift;
	v_label_t child_label;
	v_code *oldcp;	
	int pid;
	Atom child;


        /* Load value into src_r. */
        compile_msgld(&a->ir, false);
        v_nuke_nop();

        /*
         *  (*DANGER*): if save-shift-state could be a nop have to remove
         * nuke-nop.
         */
        save_shift_state(a->ir.level);

        /* shift src by the required amount. */
        if((shift = a->ir.u.shift.shift))
                v_lshui(src_r, src_r, shift);

        /*
         * Subtract off the loaded value from nbytes & add it to the message
         * ptr.   This, of course, can be reduced to a single, but it would
         * make checking message offsets (ala eq and ht's) more expensive.
         */
        v_subu(nbytes_r, nbytes_r, src_r);
        v_addp(msg_r, msg_r, src_r);

	if((child = a->kids.lh_first))
		child_label = v_genlabel();

	/* TRUE: fallthrough. */
	if(!(pid = a->pid)) {
		/* 
	 	 * If not a short filter, load failure reg with 
	 	 * pc of restoration code.
	 	 */
        	v_label_t restore = v_genlabel();
		int offset;

		v_setv(fail_r, restore);
		v_jv(child_label);
		offset = KID_FROM_CODE - (v_ip - &a->code[0]);

      		/* Only get here on failure. */
      		v_label(restore);
			v_raw_load(v_ldpi(fail_r, fail_r, offset),1);
            		restore_shift_state(a->ir.level);  
			v_jp(fail_r); /* jump to next node. */
	} else {
		if(!child)
			v_retii(pid);
		else {
			v_label_t pid_label = v_genlabel();
			v_setv(fail_r, pid_label);
			v_jv(child_label);
			v_label(pid_label);
				v_retii(pid);
		}
	}

	if(child) {
                oldcp = v_swapcp(child->code);
                v_label(child_label);
                v_swapcp(oldcp);
	}
}



/* Generate code for hte entry - hard-code jumps. */
static void compile_hte(Atom hte) {
        int pid;

	v_begin_incremental(hte->code, sizeof hte->code);
        pid = hte->pid;

        /* If no kids, just return id. */
        if(!hte->kids.lh_first) {
                v_retii(pid);
	/* expected case */
	} else if(!pid) {
		/* jump to successor */
		v_ldpi(jump_r, jump_r, KID_FROM_CODE);
		v_addpi(jump_r, jump_r, CODE_OFFSET);
		v_jp(jump_r);
	/* Short filter: load failure reg with address to jump too. */
	} else {
                v_label_t l;

               	l = v_genlabel();
                v_setv(fail_r, l);

		v_ldpi(jump_r, jump_r, KID_FROM_CODE);
		v_addpi(jump_r, jump_r, CODE_OFFSET);
		v_jp(jump_r);

		/* If the atoms for this node reject, jump here. */
		v_label(l);
                       	v_retii(pid);
	}
	v_end_incremental();
}

/*
 * if(hte->val != val)
 *      goto out;
 * else
 *      goto hte->label;
 */
static void gen_lookup(Atom a, v_label_t out) {
	unsigned state;

        v_raw_load(v_ldui(val_r, hte_r, offsetof(struct atom, ir.u.eq.val)), 1);

        v_addpi(jump_r, hte_r, CODE_OFFSET);
        v_bnep(src_r, val_r, out);      /* hte->val == val */

	state = a->ht->state;

	/* else load pid and return. */
	if(state & DPF_ALL_TERMS) {
		switch(sizeof(fid_t)) {
		case 1:
        		v_lduci(val_r, hte_r, offsetof(struct atom, pid));
			break;
		case 2: 
			v_ldusi(val_r, hte_r, offsetof(struct atom, pid));
			break;
		case 4: 
			v_ldui(val_r, hte_r, offsetof(struct atom, pid)); 
			break;
		}
		v_reti(val_r);
	} else {
		/* all non-terms: jump directly. */
		if(state & DPF_ALL_NTERMS) {
			v_ldpi(jump_r, jump_r, KID_FROM_CODE);
			v_addpi(jump_r, jump_r, CODE_OFFSET);
		}
        	v_jp(jump_r);          /* goto hte->code */
	}
}

/* 
 * Compute what special-case optimizations we can do.  Currently
 * optimize for:
 *	1. hash table has no collisions.
 *	2. hash table has only terminals.
 *	3. hash table has only non-terminals.
 */
unsigned ht_state(Ht ht) {
        unsigned state;

        state = 0;

        /* no collisions means we can elide collision checks. */
        if(!ht->coll)
                state |= DPF_NO_COLL;

        /*
         * note: when the hash table is empty, it will trigger ALL_NTERMS,
         * HOWEVER, the hash table will be set to REGEN, which will
         * make ht_regenp return 1.  A bit sleazy.  Weep.
         */
        if(!ht->nterm)
                state |= DPF_ALL_TERMS;
        else if(!ht->term)
                state |= DPF_ALL_NTERMS;

        return state;
}

/*
 * Generate code for the hash table:
         for(hte = hash(ht, val); hte != 0; hte = hte->or)
                if(hte->val == val)
                        goto hte->label;
 */
static void compile_ht(Atom a, v_label_t l) {
        v_label_t loop, elsel;
        unsigned mask, log2_hte, state;
        int prec_bits;
	struct ir *ir;
	Ht ht;

	ht = a->ht;
	state = ht->state = ht_state(ht);

	ir = &a->ir;
	compile_msgld(ir, l);
	v_nuke_nop();

        /*
         * hte = ht[(val * hashmult) >> prec_bits) & mask];
         */
        log2_hte = log2(sizeof ht->ht[0]);
        prec_bits = 32 - (ht->log2_sz + log2_hte);

        /* shift mask into the correct position. */
        mask = (ht->htsz - 1) << log2_hte;

        v_mului(val_r, src_r, hashmult);        /* compute hash function */
        v_setp(hte_r, &ht->ht[0]);                 /* load base ptr */
        v_rshui(val_r, val_r, prec_bits);
        v_andui(val_r, val_r, mask);
        v_ldp(hte_r, val_r, hte_r);                       /* Load hte */

        /* hte != 0 */
        v_beqpi(hte_r, 0, l);

        /* elide loop if there are no collisions. */
        if(state & DPF_NO_COLL) {
                gen_lookup(a, l);
	/* Unroll the collision loop to the length of the longest chain? */
        } else {
                loop = v_genlabel(); elsel = v_genlabel();

                v_label(loop);
                        gen_lookup(a, elsel);

                v_label(elsel);
                        v_ldpi(hte_r, hte_r, offsetof(struct atom, next));
			/* hte == 0, goto loop */
                        v_bnepi(hte_r, 0, loop);        
        }
}

/* Compile == expression. */
static void compile_eq(Atom a, v_label_t false) {
	struct eq *e;
	struct ir *ir;
	v_label_t true, pid_label;
	v_code *oldcp;
	int pid;
	Atom child;

	pid = a->pid;
	child = a->kids.lh_first;

	true = v_genlabel();
	demand(!a->ht, bogus: should not be a hash table);

	ir = &a->ir;
	e = &ir->u.eq;	
	compile_msgld(ir, false);

	/* fill delay slot with possibly unnecessary imm load. */
	v_nuke_nop();
	v_setu(t0, e->val);

	/* Short filter prologue. */
	if(pid && child) {
 		pid_label = v_genlabel();
		v_setv(fail_r, pid_label);
	}

	/* terminal */
	if(pid && !child) {
		v_bneu(src_r, t0, false);
		v_retii(pid);
	/* non-terminal. */
	} else {
		v_bequ(src_r, t0, true);

		/* Setup label for child code. */
		oldcp = v_swapcp(child->code);
		v_label(true);
		v_swapcp(oldcp);

		/* Short filter epilogue. */
		if(pid && child) {
			v_label(pid_label);
				v_retii(pid);
		} 
	}
}

/* 
 * Compile atom.   Gen is responsible for emitting prologue code for pids
 * and dealing with or nodes (prologue, epilogue, and general fixup).
 * It demuxes the actual atom off to the different routines.
 * 
 * Basic algorithm:
	% Compute where to jump to on failure.
	IF in or-list THEN
 		IF first in or-list THEN
			emit: save failure reg

		% Opt: elide this if short filter and success.
		IF last guy in or-list THEN
			emit: restore failure reg
		ELSE
			emit: load sibling pc into failure reg


	% Hash tables do their own jumps.
	IF is a hash table THEN
		compile_ht
	ELSE
		% Compute where to jump to on success.
		IF successor THEN
			emit: load successor pc into jump reg
	
		% Compute boolean
		[ Emit code for eq or shift - they all jump to false on failure ]

		% TRUE: fall through of generated code.
			IF has a pid THEN
				IF has no kids THEN
					emit: return pid
				ELSE
					emit: load pid label into failure reg
					emit: jump using jump reg
					emit: pid label, return of pid.
			ELSE
				emit: jump to successor (using jump reg)
	
	emit: false label
		emit: jump to failure reg	
 *
 * Need to schedule all the stuff below, of course.
 */
static void compile_atom(Atom a) {
	v_label_t false;
	Atom or;
	struct ir *ir;
 	int first_or, orlist;

	first_or = firstor_p(a);
	orlist = orlist_p(a);

	ir = &a->ir;
	or = a->sibs.le_next;

	v_begin_incremental(a->code, sizeof a->code);

	/* Compute where to jump to on failure */
	if(orlist) {
		/* if first guy, save failure reg */
		if(first_or)
			v_stpi(fail_r, v_lp, FAIL_OFFSET(ir->level));
		/* if last guy, restore failure reg */
		if(!or)
			v_ldpi(fail_r, v_lp, FAIL_OFFSET(ir->level));
		/* else load sib's pc into failure reg */
		else {
			v_setp(fail_r, &a->sibs.le_next);	
			v_ldpi(fail_r, fail_r, 0);
			v_addpi(fail_r, fail_r, CODE_OFFSET);
#if 0
			v_setp(fail_r, &a->code[0]);	
			v_ldpi(fail_r, fail_r, SIB_FROM_CODE);
			v_addpi(fail_r, fail_r, CODE_OFFSET);
#endif
		}
	}
	false = v_genlabel();

	/* weird structure, neh? */
	if(a->ht) 
		compile_ht(a, false);
	else if(isshift(&a->ir))
		compile_shift(a, false);
	else
		compile_eq(a, false);

	/* 
	 * FALSE -- note we could save on code space by creating
	 * this trampoline a single time.  The only problem would be
	 * that branches to it might frequently not fit in the branc
	 * immediate field, forcing an additional indirect jump.
	 */
	v_label(false);
		v_jp(fail_r);

	v_end_incremental();
}


void dpf_compile_atom(Atom a) {
	Atom p;

	if(a == dpf_base)
		return;
	if(!(p = a->parent))
		fatal(bogus parent);
	if(p->ht)
		compile_hte(a);
	else
		compile_atom(a);
}

/* Pointer to where demux prologue jump instruction should go. */
static v_code *jump_ip;

static v_iptr compile_first_atom(Atom base) {
	/* Memory to hold code in -- size is a hack. */
	v_reg_t args[2];
	v_label_t no_match;
	Atom trie;
	int i;
	/* maximum instructions that a jump should take. */
#	define PADDING	4	
        static v_code *insn;
	static int nbytes;

	if(!insn) {
		nbytes = 256 * sizeof *insn;
		insn = malloc(nbytes);
		demand(insn, fatal: malloc out of memory);
	}

	trie = base->kids.lh_first;

	/* 
	 * I guess we use insn as a place to store trampolines of too-large
	 * jumps, neh?  Or should be give a special place for that?  How
	 * to track which parts are being used in that way?
	 */
        v_lambda("dpf-filter", "%p%u", args, V_LEAF, insn, nbytes);
		if(!v_getreg(&src_r, V_UL, V_TEMP)
		|| !v_getreg(&fail_r, V_P, V_TEMP)
		|| !v_getreg(&jump_r, V_P, V_TEMP)
		|| !v_getreg(&t0, V_UL, V_TEMP))
			fatal(out of registers!);

        	/* Allocate hash table registers */
        	if(!v_getreg(&hte_r, V_P, V_TEMP)
        	|| !v_getreg(&val_r, V_UL, V_TEMP))
                	fatal(Out of registers);

		/* Allocate space to save fail_r. */
		fail_save_area = v_localb(sizeof(void *) * DPF_MAXFILT);

		/* Allocate space to save nbytes at every shift. */
		nbytes_save_area = v_localb(sizeof(unsigned) * DPF_MAXFILT);

		/* Allocate space to save message pointer at every shift. */
		msg_save_area = v_localb(sizeof(void *) * DPF_MAXFILT);
		
		msg_r = args[0]; 
		nbytes_r = args[1];
		no_match = v_genlabel();

		v_setv(fail_r, no_match);

		/* Patch in jump using SMC */
		for(jump_ip = v_ip, i = 0; i < PADDING; i++)
			v_nop();

#if 0
		/* Load pc of first entry and jump to it. */
		v_setp(jump_r, &base->kids.lh_first);
		v_raw_load( v_ldpi(jump_r, jump_r, 0), 1);
		v_addpi(jump_r, jump_r, CODE_OFFSET);
		v_jp(jump_r);
#endif

		v_label(no_match);	/* jump here if no matches */
			v_retii(0);
	return v_end(0).i;
}

/* Compile DPF trie.  Need to reduce register requirements. */
v_iptr dpf_compile(Atom base) {
	static v_iptr ip;
	static Atom first_atom;
	Atom a;

	a = base->kids.lh_first;

	if(!ip)
		ip = compile_first_atom(base);
	
	/* Backpatch jump to first atom's code using SMC. */
	if(a && first_atom != a) {
		v_code *tmp;

		first_atom = a;
		tmp = v_swapcp(jump_ip);
		v_jpi(a->code);
		v_flushcache(jump_ip, (v_ip - jump_ip) * sizeof *v_ip);
		v_swapcp(tmp);
	}
		
	return !a ? dpf_ret_zero : ip;
}
